from matchescu.blocking._block import Block
from matchescu.blocking._blocker import Blocker
from matchescu.blocking._lsh import LSHBlocker
from matchescu.blocking._sorted_neighborhood import SortedNeighborhoodBlocker
from matchescu.blocking._tf_idf import TfIdfBlocker

__all__ = [
    "Block",
    "Blocker",
    "LSHBlocker",
    "SortedNeighborhoodBlocker",
    "TfIdfBlocker",
]



import itertools
from abc import ABCMeta, abstractmethod

from matchescu.blocking._tokenization import tokenize_reference
from matchescu.reference_store.id_table._in_memory import InMemoryIdTable
from matchescu.typing import EntityReferenceIdentifier, EntityReference


class ComparisonFilter(metaclass=ABCMeta):
    @abstractmethod
    def __call__(self, *ref_ids: EntityReferenceIdentifier) -> bool:
        pass


def is_cross_source_comparison(*ref_ids: EntityReferenceIdentifier) -> bool:
    sources = set(ref_id.source for ref_id in ref_ids)
    return len(sources) > 1


class JaccardSimilarityFilter(ComparisonFilter):
    def __init__(self, id_table: InMemoryIdTable, min_similarity: float = 0.5) -> None:
        self._id_table = id_table
        self._min_similarity = min_similarity

    @staticmethod
    def _similarity(ref1: EntityReference, ref2: EntityReference) -> float:
        tok1 = set(tokenize_reference(ref1))
        tok2 = set(tokenize_reference(ref2))
        n_common = len(tok1 & tok2)
        n_total = len(tok1 | tok2)
        return n_common / n_total if n_total != 0 else 0

    def __call__(self, *ref_ids: EntityReferenceIdentifier) -> bool:
        refs = map(self._id_table.get, ref_ids)
        scores = [self._similarity(*comb) for comb in itertools.combinations(refs, 2)]
        return max(scores) >= self._min_similarity



from abc import ABCMeta, abstractmethod
from typing import Generator

from matchescu.reference_store.id_table import IdTable

from matchescu.blocking._block import Block


class Blocker(metaclass=ABCMeta):
    def __init__(self, id_table: IdTable):
        self._id_table = id_table

    @abstractmethod
    def __call__(self) -> Generator[Block, None, None]:
        pass



from collections.abc import Iterable
from typing import Generator, cast

from datasketch import MinHash, MinHashLSH

from matchescu.blocking import Block
from matchescu.blocking._blocker import Blocker
from matchescu.blocking._tokenization import tokenize_reference
from matchescu.reference_store.id_table import IdTable
from matchescu.typing import EntityReference, EntityReferenceIdentifier


class LSHBlocker(Blocker):
    def __init__(
        self,
        id_table: IdTable,
        threshold: float = 0.5,
        num_perm: int = 128,
        bands: int = 16,
    ):
        super().__init__(id_table)
        self.__sim_threshold = threshold  # Similarity threshold for LSH
        self.__num_perm = num_perm  # Number of permutations for MinHash
        self.__num_bands = bands  # Number of bands for LSH bucketing
        self.__lsh = MinHashLSH(
            threshold=self.__sim_threshold, num_perm=self.__num_perm
        )

    def __compute_minhash(self, tokens: Iterable[str]) -> MinHash:
        """Compute MinHash signature for a set of tokens."""
        m = MinHash(num_perm=self.__num_perm)
        m.update_batch(map(lambda t: t.encode("utf-8"), tokens))
        return m

    def __add_entity(self, ref: EntityReference) -> None:
        """Add an entity with tokenized attributes to LSH."""
        minhash = self.__compute_minhash(tokenize_reference(ref))
        self.__lsh.insert(ref.id, minhash)

    def __call__(self) -> Generator[Block, None, None]:
        for ref in self._id_table:
            self.__add_entity(ref)

        for ref in self._id_table:
            block_key = self.__compute_minhash(tokenize_reference(ref))
            neighbor_ids = self.__lsh.query(block_key)
            yield Block(key=block_key).extend(
                cast(EntityReferenceIdentifier, x) for x in neighbor_ids
            )
